;==============================================================================

; file: "_parms.scm"

; Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved.

(include "fixnum.scm")

;------------------------------------------------------------------------------
;
; Compiler parameters module
; --------------------------

; This module contains definitions that parameterize the behavior of
; the compiler.

;------------------------------------------------------------------------------

; Symbol "canonicalization".

; (string->canonical-symbol str) behaves like 'string->symbol' but all
; the letters in the symbol are converted to the default case.

(define (string->canonical-symbol str)
  (let* ((len (string-length str))
         (s (make-string len)))
    (let loop ((i (- len 1)))
      (if (>= i 0)
        (begin
          (string-set! s i (convert-case (string-ref str i)))
          (loop (- i 1)))
        (string->symbol s)))))


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Specific symbols needed by the compiler.

(define QUOTE-sym              (string->canonical-symbol "quote"))
(define QUASIQUOTE-sym         (string->canonical-symbol "quasiquote"))
(define UNQUOTE-sym            (string->canonical-symbol "unquote"))
(define UNQUOTE-SPLICING-sym   (string->canonical-symbol "unquote-splicing"))
(define LAMBDA-sym             (string->canonical-symbol "lambda"))
(define IF-sym                 (string->canonical-symbol "if"))
(define SET!-sym               (string->canonical-symbol "set!"))
(define COND-sym               (string->canonical-symbol "cond"))
(define =>-sym                 (string->canonical-symbol "=>"))
(define ELSE-sym               (string->canonical-symbol "else"))
(define AND-sym                (string->canonical-symbol "and"))
(define OR-sym                 (string->canonical-symbol "or"))
(define CASE-sym               (string->canonical-symbol "case"))
(define LET-sym                (string->canonical-symbol "let"))
(define LET*-sym               (string->canonical-symbol "let*"))
(define LETREC-sym             (string->canonical-symbol "letrec"))
(define BEGIN-sym              (string->canonical-symbol "begin"))
(define DO-sym                 (string->canonical-symbol "do"))
(define DEFINE-sym             (string->canonical-symbol "define"))
(define DELAY-sym              (string->canonical-symbol "delay"))
(define FUTURE-sym             (string->canonical-symbol "future"))
(define **DEFINE-MACRO-sym     (string->canonical-symbol "##define-macro"))
(define **DECLARE-sym          (string->canonical-symbol "##declare"))
(define **INCLUDE-sym          (string->canonical-symbol "##include"))

(define NOT-sym                (string->canonical-symbol "not"))

(define **NOT-sym              (string->canonical-symbol "##not"))
(define **EQ?-sym              (string->canonical-symbol "##eq?"))
(define **EQV?-sym             (string->canonical-symbol "##eqv?"))
(define **QUASI-APPEND-sym     (string->canonical-symbol "##quasi-append"))
(define **QUASI-LIST-sym       (string->canonical-symbol "##quasi-list"))
(define **QUASI-CONS-sym       (string->canonical-symbol "##quasi-cons"))
(define **QUASI-LIST->VECTOR-sym (string->canonical-symbol "##quasi-list->vector"))
(define **MAKE-CELL-sym        (string->canonical-symbol "##make-cell"))
(define **CELL-REF-sym         (string->canonical-symbol "##cell-ref"))
(define **CELL-SET!-sym        (string->canonical-symbol "##cell-set!"))
(define **MAKE-PLACEHOLDER-sym (string->canonical-symbol "##make-placeholder"))

(define IEEE-SCHEME-sym        (string->canonical-symbol "ieee-scheme"))
(define R4RS-SCHEME-sym        (string->canonical-symbol "r4rs-scheme"))
(define MULTILISP-sym          (string->canonical-symbol "multilisp"))

(define LAMBDA-LIFT-sym        (string->canonical-symbol "lambda-lift"))

(define INLINE-sym             (string->canonical-symbol "inline"))
(define INLINING-LIMIT-sym     (string->canonical-symbol "inlining-limit"))

(define BLOCK-sym              (string->canonical-symbol "block"))
(define SEPARATE-sym           (string->canonical-symbol "separate"))

(define STANDARD-BINDINGS-sym  (string->canonical-symbol "standard-bindings"))
(define EXTENDED-BINDINGS-sym  (string->canonical-symbol "extended-bindings"))

(define SAFE-sym               (string->canonical-symbol "safe"))

(define INTERRUPTS-ENABLED-sym (string->canonical-symbol "interrupts-enabled"))

(define GENERIC-sym            (string->canonical-symbol "generic"))
(define FIXNUM-sym             (string->canonical-symbol "fixnum"))
(define FLONUM-sym             (string->canonical-symbol "flonum"))

(define NAMESPACE-sym          (string->canonical-symbol "namespace"))

(define C-DEFINE-TYPE-sym      (string->canonical-symbol "c-define-type"))
(define C-DECLARE-sym          (string->canonical-symbol "c-declare"))
(define C-INITIALIZE-sym       (string->canonical-symbol "c-initialize"))
(define C-LAMBDA-sym           (string->canonical-symbol "c-lambda"))
(define C-DEFINE-sym           (string->canonical-symbol "c-define"))

(define VOID-sym               (string->canonical-symbol "void"))
(define CHAR-sym               (string->canonical-symbol "char"))
(define SIGNED-CHAR-sym        (string->canonical-symbol "signed-char"))
(define UNSIGNED-CHAR-sym      (string->canonical-symbol "unsigned-char"))
(define LATIN1-sym             (string->canonical-symbol "latin1"))
(define UCS4-sym               (string->canonical-symbol "ucs4"))
(define UCS2-sym               (string->canonical-symbol "ucs2"))
(define SHORT-sym              (string->canonical-symbol "short"))
(define UNSIGNED-SHORT-sym     (string->canonical-symbol "unsigned-short"))
(define INT-sym                (string->canonical-symbol "int"))
(define UNSIGNED-INT-sym       (string->canonical-symbol "unsigned-int"))
(define LONG-sym               (string->canonical-symbol "long"))
(define UNSIGNED-LONG-sym      (string->canonical-symbol "unsigned-long"))
(define FLOAT-sym              (string->canonical-symbol "float"))
(define DOUBLE-sym             (string->canonical-symbol "double"))
(define STRUCT-sym             (string->canonical-symbol "struct"))
(define UNION-sym              (string->canonical-symbol "union"))
(define POINTER-sym            (string->canonical-symbol "pointer"))
(define FUNCTION-sym           (string->canonical-symbol "function"))
(define BOOLEAN-sym            (string->canonical-symbol "boolean"))
(define CHAR-STRING-sym        (string->canonical-symbol "char-string"))
(define UCS4-STRING-sym        (string->canonical-symbol "ucs4-string"))
(define UCS2-STRING-sym        (string->canonical-symbol "ucs2-string"))
(define UTF8-STRING-sym        (string->canonical-symbol "utf8-string"))
(define SCHEME-OBJECT-sym      (string->canonical-symbol "scheme-object"))


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Identifier prefix for C identifiers generated by the C-interface and
; the C back-end.

(define c-id-prefix "___")


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Identifier prefix for module names.

(define module-prefix ";")


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; C types available to the C-interface.

(define scheme-to-c-notation
  (list
    (list VOID-sym           "VOID"       "void")
    (list CHAR-sym           "CHAR"       "char")
    (list SIGNED-CHAR-sym    "SCHAR"      "signed char")
    (list UNSIGNED-CHAR-sym  "UCHAR"      "unsigned char")
    (list LATIN1-sym         "LATIN1"     (string-append c-id-prefix "LATIN1"))
    (list UCS4-sym           "UCS4"       (string-append c-id-prefix "UCS4"))
    (list UCS2-sym           "UCS2"       (string-append c-id-prefix "UCS2"))
    (list SHORT-sym          "SHORT"      "short")
    (list UNSIGNED-SHORT-sym "USHORT"     "unsigned short")
    (list INT-sym            "INT"        "int")
    (list UNSIGNED-INT-sym   "UINT"       "unsigned int")
    (list LONG-sym           "LONG"       "long")
    (list UNSIGNED-LONG-sym  "ULONG"      "unsigned long")
    (list FLOAT-sym          "FLOAT"      "float")
    (list DOUBLE-sym         "DOUBLE"     "double")
    (list BOOLEAN-sym        "BOOLEAN"    "int")
    (list CHAR-STRING-sym    "CHARSTRING" "char*")
    (list UTF8-STRING-sym    "UTF8STRING" (string-append c-id-prefix
                                                         "UTF8STRING"))
    (list UCS4-STRING-sym    "UCS4STRING" (string-append c-id-prefix
                                                         "UCS4STRING"))
    (list UCS2-STRING-sym    "UCS2STRING" (string-append c-id-prefix
                                                         "UCS2STRING"))
    (list SCHEME-OBJECT-sym  "SCMOBJ"     (string-append c-id-prefix
                                                         "WORD"))))


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; For '_source.scm'.

; Filename extensions to try in order to find source files from module name.

(define scm-file-exts '(".scm" "")) ; "" means no extension


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Version of the Gambit compiler and "gambit.h" header file.

(define compiler-version "2.5.1")
(define header-version 17)


;==============================================================================
