;==============================================================================

; file: "_t-c-2.scm"

; Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved.

(include "fixnum.scm")

;------------------------------------------------------------------------------
;
; Back end for C language (part 2)
; -----------------------

(define (targ-scan-procedure obj)
  (let* ((proc (car obj))
         (p (cdr obj))
         (nb-res (+ targ-nb-gvm-regs targ-nb-non-reg-res)))

    (if targ-info-port
      (begin
        (display "  #<" targ-info-port)
        (if (proc-obj-primitive? proc)
          (display "primitive " targ-info-port)
          (display "procedure " targ-info-port))
        (display (proc-obj-name proc) targ-info-port)
        (display ">" targ-info-port)))

    (set! targ-proc-name               (proc-obj-name proc))
    (set! targ-proc-code               (make-stretchable-vector #f))
    (set! targ-proc-code-length        0)
    (set! targ-proc-rd-res             (make-vector nb-res #f))
    (set! targ-proc-wr-res             (make-vector nb-res #f))
    (set! targ-proc-lbl-tbl            (queue-empty))
    (set! targ-proc-lbl-tbl-ord        (queue-empty))
    (set! targ-var-descr-queue         (queue-empty))
    (set! targ-first-class-label-queue (queue-empty))

    (let ((x (proc-obj-code proc)))
      (if (bbs? x)
        (targ-scan-scheme-procedure x)
        (targ-scan-c-procedure x)))

    (targ-cell-set! (caddr p) (+ targ-lbl-alloc 1))

    ; Assign label numbers sequentially, starting with "value" labels
    ; and then "goto" labels

    (let ((ord-lbls (queue->list targ-proc-lbl-tbl-ord)))
      (let loop2 ((l ord-lbls) (i 0) (val-lbls '()))
        (if (pair? l)
          (let ((x (car l)))
            (if (targ-lbl-val? x)
              (begin
                (targ-cell-set! (targ-lbl-num x) i)
                (loop2 (cdr l) (+ i 1) (cons x val-lbls)))
              (loop2 (cdr l) i val-lbls)))
          (let ((info (targ-debug-info)))
            (targ-use-obj info)
            (set! targ-lbl-alloc (+ targ-lbl-alloc (+ i 1)))
            (set-car! p
              (vector targ-proc-code
                      (reverse val-lbls)
                      targ-proc-rd-res
                      targ-proc-wr-res
                      info))
            (let loop3 ((l ord-lbls) (i i))
              (if (pair? l)
                (let ((x (car l)))
                  (if (and (targ-lbl-goto? x)
                           (not (targ-lbl-val? x)))
                    (begin
                      (targ-cell-set! (targ-lbl-num x) i)
                      (loop3 (cdr l) (+ i 1)))
                    (loop3 (cdr l) i)))))))))

    (if targ-info-port
      (newline targ-info-port))

    ))

(define (targ-debug-info)

  (define (number i lst)
    (if (null? lst)
      '()
      (cons (lst->vector (cons i (car lst)))
            (number (+ i 1) (cdr lst)))))

  (if targ-debug-info?
    (vector (lst->vector (number 0 (queue->list targ-first-class-label-queue)))
            (lst->vector (queue->list targ-var-descr-queue)))
    #f))

(define (targ-scan-scheme-procedure bbs)

  (set! targ-proc-entry-lbl   (bbs-entry-lbl-num bbs))
  (set! targ-proc-lbl-counter (make-counter (bbs-next-lbl-num bbs)))

  (let loop ((prev-bb #f)
             (prev-gvm-instr #f)
             (l (bbs->code-list bbs)))
    (if (not (null? l))
      (let ((pres-bb (code-bb (car l)))
            (pres-gvm-instr (code-gvm-instr (car l)))
            (pres-slots-needed (code-slots-needed (car l)))
            (next-gvm-instr (if (null? (cdr l))
                              #f
                              (code-gvm-instr (cadr l)))))

        (targ-gen-gvm-instr prev-gvm-instr
                            pres-gvm-instr
                            next-gvm-instr
                            pres-slots-needed)

        (loop pres-bb pres-gvm-instr (cdr l))))))

(define (targ-scan-c-procedure c-proc)

  (define (ps-opnd opnd)
    (cond ((reg? opnd)
           (let ((n (reg-num opnd)))
             (cons 'PSR n)))
          (else ; must be stack slot
           (list "PSSTK" (- (stk-num opnd) targ-proc-fp)))))

  (let* ((arity (c-proc-arity c-proc))
         (pc (targ-label-info arity #f))
         (pc-map (pcontext-map pc))
         (pc-fs (pcontext-fs pc))
         (ret (cdr (assq 'return pc-map)))
         (final-fs (+ arity 1)))

    (set! targ-proc-entry-lbl 1)

    (targ-start-bb pc-fs)

    (set! targ-proc-entry-frame targ-proc-exit-frame) ; ********** for targ-update-fr but probably not needed since it can't be called from here!
    (targ-begin-fr) ; ************* not needed either

    (targ-emit-label-entry targ-proc-entry-lbl arity #f)
    (targ-ref-lbl-goto targ-proc-entry-lbl)

    (targ-emit (list "IF_NARGS_EQ" arity '("NOTHING")))
    (targ-emit (list "WRONG_NARGS"
                     (targ-ref-lbl-val targ-proc-entry-lbl)
                     arity 0 0))
    (targ-emit (list 'APPEND
                     (list "DEF_GLBL" (targ-make-glbl "" targ-proc-name))
                     #\newline))

    (let loop1 ((i 1))
      (if (and (<= i arity) (<= i targ-nb-arg-regs))
        (begin
          (targ-emit
            (targ-loc (make-stk (+ pc-fs i)) (targ-opnd (make-reg i))))
          (loop1 (+ i 1)))))

    (targ-emit (targ-loc (make-stk final-fs) (targ-opnd ret)))
    (targ-emit (targ-adjust-stack final-fs))

    (targ-emit
      (list 'APPEND (c-proc-body c-proc)))

    (let ((adj-stk (targ-adjust-stack 0)))
      (targ-emit
        (list 'SEQ
              adj-stk
              (list "JUMPPRM"
                    '("NOTHING")
                    (targ-opnd (make-stk final-fs))))))))

;------------------------------------------------------------------------------

; Information attached to a procedure

(define targ-proc-name            #f) ; procedure's name
(define targ-proc-code            #f) ; code of the procedure
(define targ-proc-code-length     #f) ; length of code of the procedure
(define targ-proc-entry-lbl       #f) ; entry label
(define targ-proc-lbl-counter     #f) ; label counter
(define targ-proc-rd-res          #f) ; set of resources read from
(define targ-proc-wr-res          #f) ; set of resources written to
(define targ-proc-lbl-tbl         #f) ; table of all labels
(define targ-proc-lbl-tbl-ord     #f) ; table of all labels ordered by def time
(define targ-proc-fp              #f) ; frame pointer
(define targ-proc-hp              #f) ; heap pointer

(define targ-proc-instr-source    #f)
(define targ-proc-entry-frame     #f)
(define targ-proc-exit-frame      #f)

; Emit a piece of code

(define (targ-emit code)
  (stretchable-vector-set! targ-proc-code targ-proc-code-length code)
  (set! targ-proc-code-length (+ targ-proc-code-length 1)))

; Emit a label

(define (targ-emit-label-simp lbl)
  (targ-emit-label lbl #f))

(define (targ-emit-label-entry lbl nb-parms label-descr)
  (targ-emit-label lbl 'ENTRY nb-parms 0))

(define (targ-emit-label-subproc lbl nb-parms nb-closed label-descr)
  (targ-emit-label lbl 'SUBPROC nb-parms nb-closed))

(define (targ-emit-label-return lbl fs link label-descr)
  (targ-emit-label lbl 'RETURN fs link))

(define (targ-emit-label-return-task lbl fs link label-descr)
  (targ-emit-label lbl 'RETURN-TASK fs link))

(define (targ-emit-label-return-internal lbl fs link label-descr)
  (targ-emit-label lbl 'RETURN-INTERNAL fs link))

; Add a label to the code stream

(define (targ-emit-label lbl class . info)
  (let ((x (targ-get-lbl lbl)))
    (targ-emit (cons 'LABEL x))
    (targ-add-label x class info)))

(define (targ-add-label lbl-struct class info)
  (vector-set! lbl-struct 2 class)
  (vector-set! lbl-struct 3 info)
  (queue-put! targ-proc-lbl-tbl-ord lbl-struct))

; Add label "n" to label table and return label object

(define (targ-get-lbl n)
  (let ((x (assq n (queue->list targ-proc-lbl-tbl))))
    (if x
      (cdr x)
      (let ((y (vector (targ-make-cell #f) ; eventual label number (set later)
                       #f                  ; used as a "goto" label?
                       #f                  ; class (not #f if "value" label)
                       #f)))               ; extra info if "value" label
        (queue-put! targ-proc-lbl-tbl (cons n y))
        y))))

(define (targ-lbl-num lbl-struct)
  (vector-ref lbl-struct 0))

; Mark a label as referenced for "value" and return eventual label number

(define (targ-ref-lbl-val n)
  (let ((x (targ-get-lbl n)))
    (targ-lbl-num x)))

(define (targ-lbl-val? lbl-struct)
  (vector-ref lbl-struct 2))

; Mark a label as target for "goto" and return eventual label number

(define (targ-ref-lbl-goto n)
  (let ((x (targ-get-lbl n)))
    (vector-set! x 1 #t)
    (targ-make-glbl (targ-lbl-num x) targ-proc-name)))

(define (targ-lbl-goto? lbl-struct)
  (vector-ref lbl-struct 1))

(define (targ-make-glbl n name)
  (list 'GLBL n name))

; To generate new, unique labels

(define (targ-new-lbl)
  (targ-proc-lbl-counter))

(define (targ-heap-reserve space fs)
  (set! targ-proc-hp (+ targ-proc-hp space))
  (if (> targ-proc-hp targ-msection-biggest)
    (targ-check-heap space fs)))

(define (targ-check-heap space fs)
  (let ((lbl (targ-new-lbl)))
    (targ-need-heap)
    (targ-update-fr)
    (targ-end-fr)
    (targ-emit (list 'SEQ
                     (targ-adjust-stack fs)
                     (list "CHECK_HEAP"
                           (targ-ref-lbl-val lbl)
                           (+ targ-msection-biggest space))))
    (targ-gen-label-return* lbl 'RETURN-INTERNAL fs)
    (set! targ-proc-hp 0)))

(define (targ-poll fs)
  (let ((lbl (targ-new-lbl)))
    (targ-rd-fp)
    (targ-update-fr)
    (targ-end-fr)
    (targ-emit (list 'SEQ
                     (targ-adjust-stack fs)
                     (list "POLL" (targ-ref-lbl-val lbl))))
    (targ-gen-label-return* lbl 'RETURN-INTERNAL fs)))

(define (targ-start-bb fs)
  (set! targ-proc-hp 0)
  (set! targ-proc-fp fs))

(define (targ-begin-fr) ; start of a floating point region
  (targ-fp-cache-init)
  (set! targ-fr-cell (targ-make-cell #f))
  (targ-emit (list 'APPEND targ-fr-cell)))

(define (targ-update-fr)
  (let* ((frame
          targ-proc-entry-frame);*****targ-proc-exit-frame
         (live
          (frame-live frame))
         (any-closed-live?
          (varset-intersects?
            live
            (list->varset (frame-closed frame)))))

    (define (live? var)
      (or (varset-member? var live)
          (and (eq? var closure-env-var) any-closed-live?)))

    (let loop1 ((i 1) (l (reverse (frame-slots frame))))
      (if (pair? l)
        (begin
          (if (live? (car l))
            (targ-fp-cache-write-if-dirty (make-stk i)))
          (loop1 (+ i 1) (cdr l)))
        (let loop2 ((i 0) (l (frame-regs frame)))
          (if (pair? l)
            (begin
              (if (live? (car l))
                (targ-fp-cache-write-if-dirty (make-reg i)))
              (loop2 (+ i 1) (cdr l)))))))))

(define (targ-end-fr) ; end of a floating point region

  (define (make-decls i)
    (if (>= i 0)
      (cons (list "D_FR" (cons 'FR i)) (make-decls (- i 1)))
      '()))

  (let ((n (targ-fp-cache-size)))
    (if (> n 0)
      (let ((decls (reverse (make-decls (- n 1)))))
        (targ-cell-set!
          targ-fr-cell
          (list 'APPEND (cons 'SEQ (cons '("BEGIN_FR") decls)) #\newline))
        (targ-emit '(APPEND ("END_FR") #\newline))))))

; Management of resources

(define targ-nb-non-reg-res 2)

(define (targ-res-name i)
  (if (< i targ-nb-non-reg-res)
    (vector-ref '#("HEAP" "FP") i)
    (string-append "R" (number->string (- i targ-nb-non-reg-res)))))

(define (targ-rd-res i)
  (vector-set! targ-module-rd-res i #t)
  (vector-set! targ-proc-rd-res i #t))

(define (targ-wr-res i)
  (targ-rd-res i)
  (vector-set! targ-module-wr-res i #t)
  (vector-set! targ-proc-wr-res i #t))

(define (targ-need-heap)
  (targ-wr-res 0))

(define (targ-rd-fp)
  (targ-rd-res 1))

(define (targ-wr-fp)
  (targ-wr-res 1))

(define (targ-rd-reg n)
  (targ-rd-res (+ n targ-nb-non-reg-res)))

(define (targ-wr-reg n)
  (targ-wr-res (+ n targ-nb-non-reg-res)))

(define (targ-use-all-res)
  (let loop ((i (- (+ targ-nb-non-reg-res targ-nb-gvm-regs) 1)))
    (if (>= i 0)
      (begin
        (targ-wr-res i)
        (loop (- i 1))))))

;------------------------------------------------------------------------------

(define (targ-add-var-descr! descr)

  (define (index x l)
    (let loop ((l l) (i 0))
      (cond ((not (pair? l))    #f)
            ((equal? (car l) x) i)
            (else               (loop (cdr l) (+ i 1))))))

  (let ((n (index descr (queue->list targ-var-descr-queue))))
    (if n
      n
      (let ((m (length (queue->list targ-var-descr-queue))))
        (queue-put! targ-var-descr-queue descr)
        m))))

(define (targ-add-first-class-label! source slots frame)
  (let loop ((i 0) (l1 slots) (l2 '()))
    (if (pair? l1)
      (let ((var (car l1)))
        (let ((x (frame-live? var frame)))
          (if (and x (or (pair? x) (not (temp-var? x))))
            (let ((descr-index
                    (targ-add-var-descr!
                     (if (pair? x)
                        (map (lambda (y) (targ-add-var-descr! (var-name y))) x)
                        (var-name x)))))
              (loop (+ i 1) (cdr l1) (cons (+ (* i 16384) descr-index) l2)))
            (loop (+ i 1) (cdr l1) l2))))
      (let ((label-descr (cons source l2)))
        (queue-put! targ-first-class-label-queue label-descr)
        label-descr))))

;------------------------------------------------------------------------------

(define (targ-gen-gvm-instr prev-gvm-instr gvm-instr next-gvm-instr sn)

  (set! targ-proc-instr-source
    (comment-get (gvm-instr-comment gvm-instr) 'source))
  (set! targ-proc-exit-frame
    (gvm-instr-frame gvm-instr))
  (set! targ-proc-entry-frame
    (and prev-gvm-instr (gvm-instr-frame prev-gvm-instr)))

  (case (gvm-instr-type gvm-instr)

    ((LABEL)
     (set! targ-proc-entry-frame targ-proc-exit-frame)
     (targ-start-bb (frame-size targ-proc-exit-frame))
     (case (LABEL-type gvm-instr)
       ((SIMPLE)
        (targ-gen-LABEL-SIMPLE (LABEL-lbl-num gvm-instr)
                               sn))
       ((ENTRY)
        (targ-gen-LABEL-ENTRY (LABEL-lbl-num gvm-instr)
                              (LABEL-ENTRY-nb-parms gvm-instr)
                              (LABEL-ENTRY-opts gvm-instr)
                              (LABEL-ENTRY-keys gvm-instr)
                              (LABEL-ENTRY-rest? gvm-instr)
                              (LABEL-ENTRY-closed? gvm-instr)
                              sn))
       ((RETURN)
        (targ-gen-LABEL-RETURN (LABEL-lbl-num gvm-instr)
                               sn))
       ((TASK-ENTRY)
        (targ-gen-LABEL-TASK-ENTRY (LABEL-lbl-num gvm-instr)
                                   sn))
       ((TASK-RETURN)
        (targ-gen-LABEL-TASK-RETURN (LABEL-lbl-num gvm-instr)
                                    sn))
       (else
        (compiler-internal-error
          "targ-gen-gvm-instr, unknown label type"))))

    ((APPLY)
     (targ-gen-APPLY (APPLY-prim gvm-instr)
                     (APPLY-opnds gvm-instr)
                     (APPLY-loc gvm-instr)
                     sn))

    ((COPY)
     (targ-gen-COPY (COPY-opnd gvm-instr)
                    (COPY-loc gvm-instr)
                    sn))

    ((CLOSE)
     (targ-gen-CLOSE (CLOSE-parms gvm-instr)
                     sn))

    ((IFJUMP)
     (targ-gen-IFJUMP (IFJUMP-test gvm-instr)
                      (IFJUMP-opnds gvm-instr)
                      (IFJUMP-true gvm-instr)
                      (IFJUMP-false gvm-instr)
                      (IFJUMP-poll? gvm-instr)
                      (if (and next-gvm-instr
                               (memq (LABEL-type next-gvm-instr)
                               '(SIMPLE TASK-ENTRY)))
                        (LABEL-lbl-num next-gvm-instr)
                        #f)))

    ((JUMP)
     (targ-gen-JUMP (JUMP-opnd gvm-instr)
                    (JUMP-nb-args gvm-instr)
                    (JUMP-poll? gvm-instr)
                    (JUMP-safe? gvm-instr)
                    (if (and next-gvm-instr
                             (memq (LABEL-type next-gvm-instr)
                                   '(SIMPLE TASK-ENTRY)))
                      (LABEL-lbl-num next-gvm-instr)
                      #f)))

    (else
     (compiler-internal-error
       "targ-gen-gvm-instr, unknown 'gvm-instr'" gvm-instr))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-SIMPLE lbl sn)
  (targ-emit-label-simp lbl)
  (targ-begin-fr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-ENTRY lbl nb-parms opts keys rest? closed? sn)

  (let ((label-descr (targ-add-first-class-label!
                       targ-proc-instr-source
                       '()
                       targ-proc-exit-frame)))
    (if (= lbl targ-proc-entry-lbl)
      (begin
        (targ-emit-label-entry lbl nb-parms label-descr)
        (targ-ref-lbl-val lbl)
        (targ-ref-lbl-goto lbl))
      (let ((nb-closed (length (frame-closed targ-proc-exit-frame))));******
        (targ-emit-label-subproc lbl nb-parms nb-closed label-descr))))

  (let* ((nb-parms-except-rest
          (- nb-parms (if rest? 1 0)))
         (nb-keys
          (if keys (length keys) 0))
         (nb-req-and-opt
          (- nb-parms-except-rest nb-keys))
         (nb-opts
          (length opts))
         (nb-req
          (- nb-req-and-opt nb-opts))
         (lbl*
          (targ-ref-lbl-val lbl))
         (defaults
          (append opts (map cdr (or keys '())))))

    (define (make-key-descriptor)
      (let loop ((lst1 keys) (lst2 '()))
        (if (null? lst1)
          (lst->vector (reverse lst2))
          (let ((key (car lst1)))
            (loop (cdr lst1)
                  (cons (obj-val (cdr key)) (cons (car key) lst2)))))))

    (define (dispatch-on-nb-args nb-args)
      (if (> nb-args nb-req-and-opt)

        (targ-emit
         (if keys
           (list (if rest? "GET_KEY_REST" "GET_KEY")
                 lbl* nb-req nb-opts nb-keys
                 (targ-use-obj (make-key-descriptor)))
           (list (if rest? "GET_REST" "WRONG_NARGS")
                 lbl* nb-req nb-opts nb-keys)))

        (let ((nb-stacked (max 0 (- nb-args targ-nb-arg-regs)))
              (nb-stacked* (max 0 (- nb-parms targ-nb-arg-regs))))

          (define (setup-parameter i)
            (if (<= i nb-parms)
              (let* ((rest (setup-parameter (+ i 1)))
                     (src-reg (- i nb-stacked))
                     (src (cond ((<= i nb-args)
                                 (cons 'R src-reg))
                                ((and rest? (= i nb-parms))
                                 '("NUL"))
                                (else
                                 (targ-use-obj
                                  (obj-val
                                   (list-ref defaults (- i nb-req 1))))))))
                (if (<= i nb-stacked*)
                  (begin
                    (if (<= i nb-args) (targ-rd-reg src-reg))
                    (cons (list "PUSH" src) rest))
                  (if (and (<= i nb-args) (= nb-stacked nb-stacked*))
                    rest
                    (let ((dst-reg (- i nb-stacked*)))
                      (if (<= i nb-args) (targ-rd-reg src-reg))
                      (targ-wr-reg dst-reg)
                      (cons (list 'SET-R dst-reg src) rest)))))
              '()))

          (let ((x (setup-parameter (+ nb-stacked 1))))
            (targ-emit (list "IF_NARGS_EQ"
                             nb-args
                             (if (null? x) '("NOTHING") (cons 'SEQ x)))))

          (dispatch-on-nb-args (+ nb-args 1)))))

    (dispatch-on-nb-args nb-req)

    (if (= lbl targ-proc-entry-lbl)
      (targ-emit (list 'APPEND
                       (list "DEF_GLBL" (targ-make-glbl "" targ-proc-name))
                       #\newline)))

    (targ-begin-fr)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-RETURN lbl sn)
  (targ-gen-label-return* lbl 'RETURN (frame-size targ-proc-exit-frame)))

(define (targ-gen-label-return* lbl kind fs)

  (define (adjust-slots l n)
    (cond ((= n 0) l)
          ((< n 0) (adjust-slots (cdr l) (+ n 1)))
          (else    (adjust-slots (cons empty-var l) (- n 1)))))

  (let* ((frame targ-proc-entry-frame)
         (slots (reverse (adjust-slots (frame-slots frame)
                                       (- fs (frame-size frame)))))
         (regs (frame-regs frame))
         (vars (if (eq? kind 'RETURN-INTERNAL)
                 (append slots regs)
                 slots)))
    (let ((label-descr
            (targ-add-first-class-label!
              targ-proc-instr-source
              vars
              frame)))
      (let ((link (pos-in-list ret-var vars)))
        (if link
          (let ((n (length slots)))
            (case kind
              ((RETURN)
               (targ-emit-label-return lbl n link label-descr))
              ((RETURN-TASK)
               (targ-emit-label-return-task lbl n link label-descr))
              ((RETURN-INTERNAL)
               (targ-emit-label-return-internal lbl n link label-descr))
              (else
               (compiler-internal-error
                 "targ-gen-label-return*, unknown label kind")))
            (targ-begin-fr))
          (compiler-internal-error
            "targ-gen-label-return*, no return address in frame"))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-TASK-ENTRY lbl sn)

  (targ-emit-label-simp lbl)

  (let ((fs (frame-size targ-proc-exit-frame)))
    (if (= fs 0)

      (targ-emit '("PUSH" (R . 0)))

      (begin
       (targ-emit '("ADJFP" 1))
        (let loop ((i 0))
          (if (< i fs)
            (begin
              (targ-emit (list "SET_STK" (- i) (list "STK" (- (+ i 1)))))
              (loop (+ i 1)))
            (targ-emit (list "SET_STK" (- i) '(R . 0)))))))

    (targ-emit (list "TASK_PUSH" fs))

    (targ-begin-fr)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-TASK-RETURN lbl sn)
  (let ((lbl2 (targ-new-lbl))
        (fs (+ (frame-size targ-proc-exit-frame) 1)))

    (targ-start-bb fs)

    (targ-gen-label-return* lbl 'RETURN-TASK fs)
    (targ-emit (list "TASK_POP" (targ-ref-lbl-val lbl2)))
    (targ-gen-label-return* lbl2 'RETURN fs)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-APPLY prim opnds loc sn)
  (let ((proc (proc-obj-inlinable prim)))
    (if proc
      (proc opnds loc sn)
      (compiler-internal-error
        "targ-gen-APPLY, unknown 'prim'" prim))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-COPY opnd loc sn)
  (targ-emit
    (targ-loc loc (targ-opnd opnd))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-CLOSE parms sn)

  (define (close parms* sn*)
    (if (pair? parms*)

      (let* ((parm (car parms*))
             (lbl (closure-parms-lbl parm))
             (loc (closure-parms-loc parm))
             (opnds (closure-parms-opnds parm))
             (sn** (targ-sn-opnds opnds sn*)))
        (close (cdr parms*) (targ-sn-loc loc sn**))
        (let* ((x (targ-opnd loc))
               (elements (map targ-opnd opnds)))
          (targ-emit
            (list "BEGIN_SETUP_CLO"
                  (length elements)
                  x
                  (targ-ref-lbl-val lbl)))
          (for-each (lambda (elem)
                      (targ-emit
                        (list "ADD_CLO_ELEM" elem)))
                    elements)
          (targ-emit
            (list "END_SETUP_CLO"))))

      (begin

        (targ-heap-reserve
          (apply +
                 (map (lambda (parm)
                        (targ-closure-space
                          (length (closure-parms-opnds parm))))
                      parms))
          sn*)

        (for-each (lambda (parm)
                    (let ((loc (closure-parms-loc parm))
                          (opnds (closure-parms-opnds parm)))
                      (targ-emit
                        (targ-loc loc (list "ALLOC_CLO" (length opnds))))))
                  parms))))

  (close (reverse parms) sn))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-IFJUMP test opnds true-lbl false-lbl poll? next-lbl)
  (let ((proc (proc-obj-test test)))
    (if proc

      (let* ((fs (frame-size targ-proc-exit-frame))
             (fs* (targ-sn-opnds opnds fs)))

        (if (> targ-proc-hp 0)
          (targ-check-heap 0 fs*))

        (if poll?
          (targ-poll fs*))

        (targ-update-fr)

        (let ((adj-stk (targ-adjust-stack fs)))

          (define (emit adj-stk test lbl other)
            (targ-emit
              (list 'SEQ
                    adj-stk
                    (list "IF_GOTO" test (targ-ref-lbl-goto lbl))
                    other)))

          (cond ((eqv? true-lbl next-lbl)
                 (emit adj-stk (proc #t opnds fs) false-lbl #f))
                ((eqv? false-lbl next-lbl)
                 (emit adj-stk (proc #f opnds fs) true-lbl #f))
                (else
                 (emit adj-stk (proc #f opnds fs) true-lbl
                       (list "GOTO" (targ-ref-lbl-goto false-lbl)))))

          (targ-end-fr)))

      (compiler-internal-error
        "targ-gen-IFJUMP, unknown 'test'" test))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-JUMP opnd nb-args poll? safe? next-lbl)
  (let* ((fs (frame-size targ-proc-exit-frame))
         (fs* (targ-sn-opnd opnd fs)))

    (if (> targ-proc-hp 0)
      (targ-check-heap 0 fs*))

    (if poll?
      (targ-poll fs*))

    (targ-update-fr)

    (let* ((adj-stk (targ-adjust-stack fs))
           (set-nargs (if nb-args (list "SET_NARGS" nb-args) #f)))

      (cond ((lbl? opnd)
             (let ((n (lbl-num opnd)))
               (if (and next-lbl (= next-lbl n)) ; fall through?
                 (if (or adj-stk set-nargs)
                   (targ-emit
                     (list 'SEQ
                           adj-stk
                           set-nargs)))
                 (targ-emit
                   (list 'SEQ
                         adj-stk
                         set-nargs
                         (list "GOTO" (targ-ref-lbl-goto n)))))))
            ((and (obj? opnd)
                  (proc-obj? (obj-val opnd)))
             (let* ((proc (obj-val opnd))
                    (x (targ-use-prc proc #f)))
               (if (eq? (car x) 'PRM)
                 (targ-emit
                   (list 'SEQ
                         adj-stk
                         (list "JUMPPRM"
                               set-nargs
                               x)))
                 (let ((name (proc-obj-name proc)))
                   (if (and nb-args (targ-arg-check-avoidable? proc nb-args))
                     (targ-emit
                       (list 'SEQ
                             adj-stk
                             (list "JUMPINT"
                                   set-nargs
                                   x
                                   (targ-make-glbl "" name))))
                     (targ-emit
                       (list 'SEQ
                             adj-stk
                             set-nargs
                             (list "JUMPINT"
                                   '("NOTHING")
                                   x
                                   (targ-make-glbl 0 name)))))))))
            ((glo? opnd)
             (targ-emit
               (list 'SEQ
                     adj-stk
                     (cons (begin
                             (targ-wr-reg (+ targ-nb-arg-regs 1))
                             (if safe? "JUMPGLOSAFE" "JUMPGLONOTSAFE"))
                           (cons (if nb-args set-nargs '("NOTHING"))
                                 (cdr (targ-opnd opnd)))))))
            (else
             (targ-emit
               (list 'SEQ
                     adj-stk
                     (list (if nb-args
                             (begin
                               (targ-wr-reg (+ targ-nb-arg-regs 1))
                               (if safe? "JUMPGENSAFE" "JUMPGENNOTSAFE"))
                             "JUMPPRM")
                           (if nb-args set-nargs '("NOTHING"))
                           (targ-opnd opnd))))))

      (targ-end-fr))))

(define (targ-arg-check-avoidable? proc nb-args)
  (let ((x (proc-obj-call-pat proc)))
    (if (and (pair? x) (null? (cdr x))) ; proc accepts a fixed nb of args?
      (let ((arg-count (car x)))
        (= arg-count nb-args))          ; nb of arguments = nb of parameters?
      #f)))

;------------------------------------------------------------------------------

(define (targ-opnd opnd) ; fetch GVM operand

  (if (and targ-fp-cache-enabled? (or (reg? opnd) (stk? opnd)))
    (targ-fp-cache-write-if-dirty opnd))

  (cond ((reg? opnd)
         (let ((n (reg-num opnd)))
           (targ-rd-reg n)
           (cons 'R n)))

        ((stk? opnd)
         (targ-rd-fp)
         (list "STK" (- (stk-num opnd) targ-proc-fp)))

        ((glo? opnd)
         (let ((name (glo-name opnd)))
           (list "GLO"
                 (targ-use-glo name #f)
                 (targ-c-id-glo (symbol->string name)))))

        ((clo? opnd)
         (list "CLO"
               (targ-opnd (clo-base opnd))
               (clo-index opnd)))

        ((lbl? opnd)
         (let ((n (lbl-num opnd)))
           (list "LBL" (targ-ref-lbl-val n))))

        ((obj? opnd)
         (targ-use-obj (obj-val opnd)))

        (else
         (compiler-internal-error
           "targ-opnd, unknown 'opnd'" opnd))))

(define (targ-loc loc val) ; store GVM location
  (let ((x (targ-loc-no-invalidate loc val)))

    (if (and targ-fp-cache-enabled? (or (reg? loc) (stk? loc)))
      (targ-fp-cache-invalidate loc))

    x))

(define (targ-loc-no-invalidate loc val) ; store GVM location without
                                         ; invalidating flonum cache
  (cond ((reg? loc)
         (let ((n (reg-num loc)))
           (targ-wr-reg n)
           (list 'SET-R n val)))

        ((stk? loc)
         (targ-rd-fp)
         (list "SET_STK" (- (stk-num loc) targ-proc-fp) val))

        ((glo? loc)
         (let ((name (glo-name loc)))
           (list "SET_GLO"
                 (targ-use-glo name #t)
                 (targ-c-id-glo (symbol->string name))
                 val)))

        ((clo? loc)
         (list "SET_CLO"
               (targ-opnd (clo-base loc))
               (clo-index loc)))

        (else
         (compiler-internal-error
           "targ-loc, unknown 'loc'" loc))))

(define (targ-opnd-flo opnd) ; fetch unboxed flonum GVM operand
  (if (and targ-fp-cache-enabled? (or (reg? opnd) (stk? opnd)))
    (let ((x (targ-fp-cache-probe opnd)))
      (if x
        (cons 'FR x)
        (let ((x (cons 'FR (targ-fp-cache-enter opnd #f))))
          (targ-emit
            (list "SET_FR"
                  x
                  (list "FLOUNBOX" (targ-opnd opnd))))
          x)))
    (list "FLOUNBOX" (targ-opnd opnd))))

(define (targ-loc-flo loc val) ; store unboxed flonum to GVM location
  (if (and targ-fp-cache-enabled? (or (reg? loc) (stk? loc)))
    (begin
      (targ-fp-cache-invalidate loc)
      (let ((x (cons 'FR (targ-fp-cache-enter loc #t))))
        (list "SET_FR" x val)))
    (targ-loc loc (list "FLOBOX" val))))

(define (targ-adjust-stack fs)
  (if (= targ-proc-fp fs)
    #f
    (let ((fp targ-proc-fp))
      (set! targ-proc-fp fs)
      (targ-wr-fp)
      (list "ADJFP" (- fs fp)))))

(define (targ-sn-opnd opnd sn)
  (cond ((stk? opnd)
         (max (stk-num opnd) sn))
        ((clo? opnd)
         (targ-sn-opnd (clo-base opnd) sn))
        (else
         sn)))

(define (targ-sn-opnds opnds sn)
  (if (pair? opnds)
    (targ-sn-opnd (car opnds) (targ-sn-opnds (cdr opnds) sn))
    sn))

(define (targ-sn-loc loc sn)
  (if loc
    (targ-sn-opnd loc sn)
    sn))

;------------------------------------------------------------------------------

; Floating point number cache management.

(define targ-fp-cache-enabled? #f)
(set! targ-fp-cache-enabled? #t)

(define (targ-fp-cache-init)
  (set! targ-fp-cache (vector 0 '#())))

(define (targ-fp-cache-size)
  (vector-ref targ-fp-cache 0))

(define (targ-fp-cache-write loc i)
  (targ-emit
    (targ-loc-no-invalidate loc (list "FLOBOX" (cons 'FR i)))))

(define (targ-fp-cache-write-if-dirty loc)
  (let ((v (vector-ref targ-fp-cache 1)))
    (let ((n (vector-length v)))
      (let loop ((i 0))
        (if (< i n)
          (let ((x (vector-ref v i)))
            (if (and x (cdr x) (eqv? (car x) loc))
              (begin
                (targ-fp-cache-write loc i)
                (set-cdr! x #f))
              (loop (+ i 1)))))))))

(define (targ-fp-cache-enter opnd dirty?) ; allocate new entry for opnd
  (let ((v1 (vector-ref targ-fp-cache 1)))
    (let ((n (vector-length v1)))
      (let loop1 ((i 0))
        (if (< i n)
          (if (vector-ref v1 i)
            (loop1 (+ i 1))
            (begin
              (vector-set! v1 i (cons opnd dirty?))
              i))
          (let ((v2 (make-vector (+ n 1))))
            (let loop2 ((i 0))
              (if (< i n)
                (begin
                  (vector-set! v2 i (vector-ref v1 i))
                  (loop2 (+ i 1)))))
            (vector-set! v2 n (cons opnd dirty?))
            (vector-set! targ-fp-cache 0 (+ n 1))
            (vector-set! targ-fp-cache 1 v2)
            n))))))

(define (targ-fp-cache-probe opnd) ; opnd must be a reg or stack slot
  (let ((v (vector-ref targ-fp-cache 1)))
    (let ((n (vector-length v)))
      (let loop ((i 0))
        (if (< i n)
          (let ((x (vector-ref v i)))
            (if (and x (eqv? (car x) opnd))
              i
              (loop (+ i 1))))
          #f)))))

(define (targ-fp-cache-invalidate opnd) ; opnd must be a reg or stack slot
  (let ((v (vector-ref targ-fp-cache 1)))
    (let ((n (vector-length v)))
      (let loop ((i 0))
        (if (< i n)
          (let ((x (vector-ref v i)))
            (if (and x (eqv? (car x) opnd))
              (vector-set! v i #f))
            (loop (+ i 1))))))))

;==============================================================================

; DATABASE OF PRIMITIVES

(set! targ-prim-proc-table
  (append
    (map (lambda (x)
           (cons (string->canonical-symbol (car x))
                 (apply make-proc-obj (car x) #f #t #f (cdr x))))
'(
("##c-code"  0            #t 0        (#f)   extended)
))
targ-prim-proc-table))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Procedure specialization

(define (targ-spec name specializer-maker)
  (let ((proc-name (string->canonical-symbol name)))
    (let ((proc (targ-get-prim-info name)))
      (if proc
        (proc-obj-specialize-set! proc (specializer-maker proc proc-name))
        (compiler-internal-error
          "targ-spec, unknown primitive:" name)))))

; Safe specialization

(define (targ-s name)
  (lambda (proc proc-name)
    (let ((spec (targ-get-prim-info name)))
      (lambda (decls args) spec))))

; Unsafe specialization

(define (targ-u name)
  (lambda (proc proc-name)
    (let ((spec (targ-get-prim-info name)))
      (lambda (decls args) (if (not (safe? decls)) spec proc)))))

; Safe arithmetic specialization

(define (targ-s-a fix-name flo-name)
  (targ-arith #t fix-name flo-name))

; Unsafe arithmetic specialization

(define (targ-u-a fix-name flo-name)
  (targ-arith #f fix-name flo-name))

(define (targ-arith fix-safe? fix-name flo-name)
  (lambda (proc proc-name)
    (let ((fix-spec (if fix-name (targ-get-prim-info fix-name) proc))
          (flo-spec (if flo-name (targ-get-prim-info flo-name) proc)))
      (lambda (decls args)
        (let ((arith (arith-implementation proc-name decls)))
          (cond ((eq? arith FIXNUM-sym)
                 (if (or fix-safe? (not (safe? decls))) fix-spec proc))
                ((eq? arith FLONUM-sym)
                 (if (not (safe? decls)) flo-spec proc))
                (else
                 proc)))))))

; Safe specialization for ##eqv?

(define (targ-s-**eqv?)
  (lambda (proc proc-name)
    (let ((spec (targ-get-prim-info "##eq?")))
      (lambda (decls args)
        (if (and (= (length args) 2)
                 (or (targ-eq-testable-object? (car args))
                     (targ-eq-testable-object? (cadr args))))
          spec
          proc)))))

(define (targ-eq-testable-object? obj)
  (and (not (void-object? obj)) ; the void-object denotes a non-constant
       (or (symbol-object? obj)
           (keyword-object? obj)
           (memq (targ-obj-type obj)
                 '(BOOLEAN NULL ABSENT VOID EOF OPTIONAL REST KEY SCRIPT
                   FIXNUM CHAR)))))


;------------------------------------------------------------------------------

(define (targ-op name descr)
  (descr (targ-get-prim-info name)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-APPLY-alloc compute-space flo-result? f)
  (targ-setup-inlinable-proc
    #f
    flo-result?
    (lambda (opnds sn)
      (targ-heap-reserve
        (compute-space (length opnds))
        (targ-sn-opnds opnds sn))
      (f opnds sn))))

(define (targ-APPLY-cons)
  (targ-APPLY-alloc
    (lambda (n) targ-pair-space)
    #f
    (targ-APPLY-simp-generator #f "CONS")))

(define (targ-APPLY-list)
  (targ-APPLY-alloc
    (lambda (n) (* n targ-pair-space))
    #f
    (lambda (opnds sn)
      (cond ((null? opnds)
             '("NUL"))
            ((null? (cdr opnds))
             (list "CONS" (targ-opnd (car opnds)) '("NUL")))
            (else
             (let ((rev-elements (reverse (map targ-opnd opnds))))
               (targ-emit
                 (list "BEGIN_ALLOC_LIST"
                       (length rev-elements)
                       (car rev-elements)))
               (for-each (lambda (elem)
                           (targ-emit
                             (list "ADD_LIST_ELEM" elem)))
                         (cdr rev-elements))
               (targ-emit
                 (list "END_ALLOC_LIST"))
               (list "GET_LIST" (length rev-elements))))))))

(define (targ-APPLY-make-cell)
  (targ-APPLY-alloc
    (lambda (n) targ-cell-space)
    #f
    (targ-APPLY-simp-generator #f "MAKECELL")))

(define (targ-APPLY-make-will)
  (targ-APPLY-alloc
    (lambda (n) targ-will-space)
    #f
    (targ-APPLY-simp-generator #f "MAKEWILL")))

(define (targ-APPLY-make-placeholder)
  (targ-APPLY-alloc
    (lambda (n) targ-placeholder-space)
    #f
    (targ-APPLY-simp-generator #f "MAKEPLACEHOLDER")))

(define (targ-APPLY-vector kind)

  (define (compute-space n)
    (case kind
      ((STRING)   (targ-string-space n))
      ((VECTOR8)  (targ-vector8-space n))
      ((VECTOR16) (targ-vector8-space (* n 2)))
      (else       (targ-vector-space n))))

  (define begin-allocator-name
    (case kind
      ((STRING)   "BEGIN_ALLOC_STRING")
      ((VECTOR8)  "BEGIN_ALLOC_VECTOR8")
      ((VECTOR16) "BEGIN_ALLOC_VECTOR16")
      (else       "BEGIN_ALLOC_VECTOR")))

  (define end-allocator-name
    (case kind
      ((STRING)   "END_ALLOC_STRING")
      ((VECTOR8)  "END_ALLOC_VECTOR8")
      ((VECTOR16) "END_ALLOC_VECTOR16")
      (else       "END_ALLOC_VECTOR")))

  (define add-element
    (case kind
      ((STRING)   "ADD_STRING_ELEM")
      ((VECTOR8)  "ADD_VECTOR8_ELEM")
      ((VECTOR16) "ADD_VECTOR16_ELEM")
      (else       "ADD_VECTOR_ELEM")))

  (define getter-operation
    (case kind
      ((STRING)   "GET_STRING")
      ((VECTOR8)  "GET_VECTOR8")
      ((VECTOR16) "GET_VECTOR16")
      (else       "GET_VECTOR")))

  (targ-APPLY-alloc
    compute-space
    #f
    (lambda (opnds sn)
      (let ((elements (map targ-opnd opnds)))
        (targ-emit
          (list begin-allocator-name (length elements)))
        (for-each (lambda (elem)
                    (targ-emit
                      (list add-element elem)))
                  elements)
        (targ-emit
          (list end-allocator-name))
        (list getter-operation (length elements))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-APPLY-force)
  (lambda (prim)
    (proc-obj-inlinable-set!
      prim
      (lambda (opnds loc sn)
        (let ((lbl (targ-new-lbl))
              (opnd (car opnds))
              (sn* (targ-sn-loc loc sn)))

          (targ-update-fr)
          (targ-emit (list "FORCE1"
                           (targ-ref-lbl-val lbl)
                           (targ-opnd opnd)))
          (targ-end-fr)
          (targ-gen-label-return* lbl 'RETURN-INTERNAL sn*)
          (targ-emit (list "FORCE2"))
          (if loc
            (targ-emit (targ-loc loc (list "FORCE3")))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-APPLY-identity)
  (targ-setup-inlinable-proc*
    (lambda (opnds sn)
      (targ-opnd (car opnds)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-IFJUMP-simp name)
  (targ-setup-test-proc*
    (targ-IFJUMP-simp-generator #f name)))

(define (targ-IFJUMP-simpflo name)
  (targ-setup-test-proc*
    (targ-IFJUMP-simp-generator #t name)))

(define (targ-IFJUMP-fold name)
  (targ-setup-test-proc*
    (targ-IFJUMP-fold-generator #f name)))

(define (targ-IFJUMP-foldflo name)
  (targ-setup-test-proc*
    (targ-IFJUMP-fold-generator #t name)))

(define (targ-IFJUMP-APPLY name)
  (targ-setup-inlinable-proc*
    (targ-APPLY-simp-generator #f name)))

(define (targ-APPLY-simp side-effects? name)
  (targ-setup-inlinable-proc
    side-effects?
    #f
    (targ-APPLY-simp-generator #f name)))

(define (targ-APPLY-fold name0 name1 name2)
  (targ-setup-inlinable-proc
    #f
    #f
    (targ-APPLY-fold-generator name0 name1 name2 #f)))

(define (targ-APPLY-simpflo* name) ; primitive with flonum result
  (targ-APPLY-alloc
    (lambda (n) targ-flonum-space)
    #t
    (targ-APPLY-simp-generator #f name)))

(define (targ-APPLY-simpflo name) ; primitive with flonum args and result
  (targ-APPLY-alloc
    (lambda (n) targ-flonum-space)
    #t
    (targ-APPLY-simp-generator #t name)))

(define (targ-APPLY-foldflo name0 name1 name2) ; primitive with flonum args and
  (targ-APPLY-alloc                            ; result
    (lambda (n) targ-flonum-space)
    #t
    (targ-APPLY-fold-generator name0 name1 name2 #t)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-setup-test-proc* generator)
  (lambda (prim)
    ((targ-setup-test-proc generator)
     prim)
    ((targ-setup-inlinable-proc
       #f
       #f
       (lambda (opnds sn)
         (list "BOOL" (generator opnds sn))))
     prim)))

(define (targ-setup-test-proc generator)
  (lambda (prim)
    (proc-obj-test-set!
      prim
      (lambda (not? opnds fs)
        (let ((test (generator opnds fs)))
          (if not?
            (list "NOT" test)
            test))))))

(define (targ-IFJUMP-simp-generator flo? name)
  (lambda (opnds fs)
    (targ-IFJUMP-simp-gen opnds flo? name)))

(define (targ-IFJUMP-simp-gen opnds flo? name)
  (let loop ((l opnds) (args '()))
    (if (pair? l)
      (let ((opnd (car l)))
        (loop (cdr l)
              (cons (if flo? (targ-opnd-flo opnd) (targ-opnd opnd))
                    args)))
      (cons name (reverse args)))))

(define (targ-IFJUMP-fold-generator flo? name)
  (lambda (opnds fs)
    (targ-IFJUMP-fold-gen opnds flo? name)))

(define (targ-IFJUMP-fold-gen opnds flo? name)

  (define (multi-opnds opnds)
    (let* ((opnd1 (car opnds))
           (opnd2 (cadr opnds))
           (opnd1* (if flo? (targ-opnd-flo opnd1) (targ-opnd opnd1)))
           (opnd2* (if flo? (targ-opnd-flo opnd2) (targ-opnd opnd2)))
           (r (list name opnd1* opnd2*)))
      (if (pair? (cddr opnds))
        (list "AND" r (multi-opnds (cdr opnds)))
        r)))

  (cond ((or (not (pair? opnds))
             (not (pair? (cdr opnds))))
         1)
        (else
         (multi-opnds opnds))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-setup-inlinable-proc* generator)
  (lambda (prim)
    ((targ-setup-test-proc
       (lambda (opnds fs)
         (list "NOT" (list "FALSEP" (generator opnds fs)))))
     prim)
    ((targ-setup-inlinable-proc #f #f generator)
     prim)))

(define (targ-setup-inlinable-proc side-effects? flo-result? generator)
  (lambda (prim)
    (proc-obj-inlinable-set!
      prim
      (lambda (opnds loc sn)
        (if side-effects? ; always generate code for side-effecting operations
          (let ((x (generator opnds sn)))
            (targ-emit ; if a result is needed, copy the first argument
              (if (and loc (not (eqv? (car opnds) loc)))
                (list 'SEQ
                      x
                      (targ-loc loc (targ-opnd (car opnds))))
                x)))
          (if loc ; only generate functional operations if result needed
            (let ((sn* (targ-sn-loc loc sn)))
              (let ((x (generator opnds sn*)))
                (targ-emit
                  (if flo-result?
                    (targ-loc-flo loc x)
                    (targ-loc loc x)))))))))))

(define (targ-APPLY-simp-generator flo? name)
  (lambda (opnds sn)
    (targ-APPLY-simp-gen opnds flo? name)))

(define (targ-APPLY-simp-gen opnds flo? name)
  (let loop ((l opnds) (args '()))
    (if (pair? l)
      (let ((opnd (car l)))
        (loop (cdr l)
              (cons (if flo? (targ-opnd-flo opnd) (targ-opnd opnd))
                    args)))
      (cons name (reverse args)))))

(define (targ-APPLY-fold-generator name0 name1 name2 flo?)
  (lambda (opnds sn)
    (targ-APPLY-fold-gen opnds name0 name1 name2 flo?)))

(define (targ-APPLY-fold-gen opnds name0 name1 name2 flo?)
  (if (not (pair? opnds))
    name0
    (let* ((o (car opnds))
           (r (if flo? (targ-opnd-flo o) (targ-opnd o))))
      (if (not (pair? (cdr opnds)))
        (list name1 r)
        (let loop ((l (cdr opnds)) (r r))
          (if (pair? l)
            (let ((opnd (car l)))
              (loop (cdr l)
                    (list name2
                          r
                          (if flo? (targ-opnd-flo opnd) (targ-opnd opnd)))))
            r))))))

;------------------------------------------------------------------------------

; Table of inlinable operations (for APPLY and IFJUMP gvm instructions)

(define (targ-setup-operations)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##type"                (targ-APPLY-simp #f "TYPE"))
(targ-op "##type-cast"           (targ-APPLY-simp #f "TYPECAST"))
(targ-op "##subtype"             (targ-APPLY-simp #f "SUBTYPE"))
(targ-op "##subtype-set!"        (targ-APPLY-simp #t "SUBTYPESET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##not"                 (targ-IFJUMP-simp "FALSEP"))
(targ-op "##null?"               (targ-IFJUMP-simp "NULLP"))
(targ-op "##unbound?"            (targ-IFJUMP-simp "UNBOUNDP"))
(targ-op "##eq?"                 (targ-IFJUMP-simp "EQP"))
(targ-op "##eof-object?"         (targ-IFJUMP-simp "EOFP"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##fixnum?"             (targ-IFJUMP-simp "FIXNUMP"))
(targ-op "##flonum?"             (targ-IFJUMP-simp "FLONUMP"))
(targ-op "##special?"            (targ-IFJUMP-simp "SPECIALP"))
(targ-op "##pair?"               (targ-IFJUMP-simp "PAIRP"))
(targ-op "##subtyped?"           (targ-IFJUMP-simp "SUBTYPEDP"))
(targ-op "##procedure?"          (targ-IFJUMP-simp "PROCEDUREP"))
(targ-op "##placeholder?"        (targ-IFJUMP-simp "PLACEHOLDERP"))
(targ-op "##vector?"             (targ-IFJUMP-simp "VECTORP"))
(targ-op "##symbol?"             (targ-IFJUMP-simp "SYMBOLP"))
(targ-op "##keyword?"            (targ-IFJUMP-simp "KEYWORDP"))
(targ-op "##port?"               (targ-IFJUMP-simp "PORTP"))
(targ-op "##ratnum?"             (targ-IFJUMP-simp "RATNUMP"))
(targ-op "##cpxnum?"             (targ-IFJUMP-simp "CPXNUMP"))
(targ-op "##structure?"          (targ-IFJUMP-simp "STRUCTUREP"))
(targ-op "##string?"             (targ-IFJUMP-simp "STRINGP"))
(targ-op "##bignum?"             (targ-IFJUMP-simp "BIGNUMP"))
(targ-op "##char?"               (targ-IFJUMP-simp "CHARP"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##fixnum.+"            (targ-APPLY-fold "FIX0"  "FIXPOS" "FIXADD"))
(targ-op "##fixnum.*"            (targ-APPLY-fold "FIX1"  "FIXPOS" "FIXMUL"))
(targ-op "##fixnum.-"            (targ-APPLY-fold #f      "FIXNEG" "FIXSUB"))
(targ-op "##fixnum.quotient"     (targ-APPLY-fold #f      "FIXINV" "FIXQUO"))
(targ-op "##fixnum.remainder"    (targ-APPLY-simp #f "FIXREM"))
(targ-op "##fixnum.modulo"       (targ-APPLY-simp #f "FIXMOD"))
(targ-op "##fixnum.logior"       (targ-APPLY-fold "FIX0"  "FIXPOS" "FIXIOR"))
(targ-op "##fixnum.logxor"       (targ-APPLY-fold "FIX0"  "FIXPOS" "FIXXOR"))
(targ-op "##fixnum.logand"       (targ-APPLY-fold "FIXM1" "FIXPOS" "FIXAND"))
(targ-op "##fixnum.lognot"       (targ-APPLY-simp #f "FIXNOT"))
(targ-op "##fixnum.ashr"         (targ-APPLY-simp #f "FIXASHR"))
(targ-op "##fixnum.lshr"         (targ-APPLY-simp #f "FIXLSHR"))
(targ-op "##fixnum.shl"          (targ-APPLY-simp #f "FIXSHL"))

(targ-op "##fixnum.zero?"        (targ-IFJUMP-simp "FIXZEROP"))
(targ-op "##fixnum.positive?"    (targ-IFJUMP-simp "FIXPOSITIVEP"))
(targ-op "##fixnum.negative?"    (targ-IFJUMP-simp "FIXNEGATIVEP"))
(targ-op "##fixnum.odd?"         (targ-IFJUMP-simp "FIXODDP"))
(targ-op "##fixnum.even?"        (targ-IFJUMP-simp "FIXEVENP"))
(targ-op "##fixnum.="            (targ-IFJUMP-fold "FIXEQ"))
(targ-op "##fixnum.<"            (targ-IFJUMP-fold "FIXLT"))
(targ-op "##fixnum.>"            (targ-IFJUMP-fold "FIXGT"))
(targ-op "##fixnum.<="           (targ-IFJUMP-fold "FIXLE"))
(targ-op "##fixnum.>="           (targ-IFJUMP-fold "FIXGE"))

(targ-op "##fixnum.->char"       (targ-APPLY-simp #f "FIXTOCHR"))
(targ-op "##fixnum.<-char"       (targ-APPLY-simp #f "FIXFROMCHR"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##flonum.->fixnum"     (targ-APPLY-simp #f "FLOTOFIX"))
(targ-op "##flonum.<-fixnum"     (targ-APPLY-simpflo* "FLOFROMFIX"))
(targ-op "##flonum.+"            (targ-APPLY-foldflo "FLO0" "FLOPOS" "FLOADD"))
(targ-op "##flonum.*"            (targ-APPLY-foldflo "FLO1" "FLOPOS" "FLOMUL"))
(targ-op "##flonum.-"            (targ-APPLY-foldflo #f     "FLONEG" "FLOSUB"))
(targ-op "##flonum./"            (targ-APPLY-foldflo #f     "FLOINV" "FLODIV"))
(targ-op "##flonum.abs"          (targ-APPLY-simpflo "FLOABS"))
(targ-op "##flonum.floor"        (targ-APPLY-simpflo "FLOFLOOR"))
(targ-op "##flonum.ceiling"      (targ-APPLY-simpflo "FLOCEILING"))
(targ-op "##flonum.round"        (targ-APPLY-simpflo "FLOROUND"))
(targ-op "##flonum.exp"          (targ-APPLY-simpflo "FLOEXP"))
(targ-op "##flonum.log"          (targ-APPLY-simpflo "FLOLOG"))
(targ-op "##flonum.sin"          (targ-APPLY-simpflo "FLOSIN"))
(targ-op "##flonum.cos"          (targ-APPLY-simpflo "FLOCOS"))
(targ-op "##flonum.tan"          (targ-APPLY-simpflo "FLOTAN"))
(targ-op "##flonum.asin"         (targ-APPLY-simpflo "FLOASIN"))
(targ-op "##flonum.acos"         (targ-APPLY-simpflo "FLOACOS"))
(targ-op "##flonum.atan"         (targ-APPLY-simpflo "FLOATAN"))
(targ-op "##flonum.sqrt"         (targ-APPLY-simpflo "FLOSQRT"))

(targ-op "##flonum.zero?"        (targ-IFJUMP-simpflo "FLOZEROP"))
(targ-op "##flonum.negative?"    (targ-IFJUMP-simpflo "FLONEGATIVEP"))
(targ-op "##flonum.positive?"    (targ-IFJUMP-simpflo "FLOPOSITIVEP"))
(targ-op "##flonum.="            (targ-IFJUMP-foldflo "FLOEQ"))
(targ-op "##flonum.<"            (targ-IFJUMP-foldflo "FLOLT"))
(targ-op "##flonum.>"            (targ-IFJUMP-foldflo "FLOGT"))
(targ-op "##flonum.<="           (targ-IFJUMP-foldflo "FLOLE"))
(targ-op "##flonum.>="           (targ-IFJUMP-foldflo "FLOGE"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##char=?"              (targ-IFJUMP-fold "CHAREQP"))
(targ-op "##char<?"              (targ-IFJUMP-fold "CHARLTP"))
(targ-op "##char>?"              (targ-IFJUMP-fold "CHARGTP"))
(targ-op "##char<=?"             (targ-IFJUMP-fold "CHARLEP"))
(targ-op "##char>=?"             (targ-IFJUMP-fold "CHARGEP"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##cons"                (targ-APPLY-cons))
(targ-op "##set-car!"            (targ-APPLY-simp #t "SETCAR"))
(targ-op "##set-cdr!"            (targ-APPLY-simp #t "SETCDR"))
(targ-op "##car"                 (targ-IFJUMP-APPLY "CAR"))
(targ-op "##cdr"                 (targ-IFJUMP-APPLY "CDR"))
(targ-op "##caar"                (targ-IFJUMP-APPLY "CAAR"))
(targ-op "##cadr"                (targ-IFJUMP-APPLY "CADR"))
(targ-op "##cdar"                (targ-IFJUMP-APPLY "CDAR"))
(targ-op "##cddr"                (targ-IFJUMP-APPLY "CDDR"))
(targ-op "##caaar"               (targ-IFJUMP-APPLY "CAAAR"))
(targ-op "##caadr"               (targ-IFJUMP-APPLY "CAADR"))
(targ-op "##cadar"               (targ-IFJUMP-APPLY "CADAR"))
(targ-op "##caddr"               (targ-IFJUMP-APPLY "CADDR"))
(targ-op "##cdaar"               (targ-IFJUMP-APPLY "CDAAR"))
(targ-op "##cdadr"               (targ-IFJUMP-APPLY "CDADR"))
(targ-op "##cddar"               (targ-IFJUMP-APPLY "CDDAR"))
(targ-op "##cdddr"               (targ-IFJUMP-APPLY "CDDDR"))
(targ-op "##caaaar"              (targ-IFJUMP-APPLY "CAAAAR"))
(targ-op "##caaadr"              (targ-IFJUMP-APPLY "CAAADR"))
(targ-op "##caadar"              (targ-IFJUMP-APPLY "CAADAR"))
(targ-op "##caaddr"              (targ-IFJUMP-APPLY "CAADDR"))
(targ-op "##cadaar"              (targ-IFJUMP-APPLY "CADAAR"))
(targ-op "##cadadr"              (targ-IFJUMP-APPLY "CADADR"))
(targ-op "##caddar"              (targ-IFJUMP-APPLY "CADDAR"))
(targ-op "##cadddr"              (targ-IFJUMP-APPLY "CADDDR"))
(targ-op "##cdaaar"              (targ-IFJUMP-APPLY "CDAAAR"))
(targ-op "##cdaadr"              (targ-IFJUMP-APPLY "CDAADR"))
(targ-op "##cdadar"              (targ-IFJUMP-APPLY "CDADAR"))
(targ-op "##cdaddr"              (targ-IFJUMP-APPLY "CDADDR"))
(targ-op "##cddaar"              (targ-IFJUMP-APPLY "CDDAAR"))
(targ-op "##cddadr"              (targ-IFJUMP-APPLY "CDDADR"))
(targ-op "##cdddar"              (targ-IFJUMP-APPLY "CDDDAR"))
(targ-op "##cddddr"              (targ-IFJUMP-APPLY "CDDDDR"))

(targ-op "##list"                (targ-APPLY-list))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##make-cell"           (targ-APPLY-make-cell))
(targ-op "##cell-ref"            (targ-IFJUMP-APPLY "CELLREF"))
(targ-op "##cell-set!"           (targ-APPLY-simp #t "CELLSET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##will?"               (targ-IFJUMP-simp "WILLP"))
(targ-op "##make-will"           (targ-APPLY-make-will))
(targ-op "##will-owner"          (targ-IFJUMP-APPLY "WILLOWNER"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##vector"              (targ-APPLY-vector 'VECTOR))
(targ-op "##vector-length"       (targ-APPLY-simp #f "VECTORLENGTH"))
(targ-op "##vector-ref"          (targ-IFJUMP-APPLY "VECTORREF"))
(targ-op "##vector-set!"         (targ-APPLY-simp #t "VECTORSET"))
(targ-op "##vector-shrink!"      (targ-APPLY-simp #t "VECTORSHRINK"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##string"              (targ-APPLY-vector 'STRING))
(targ-op "##string-length"       (targ-APPLY-simp #f "STRINGLENGTH"))
(targ-op "##string-ref"          (targ-APPLY-simp #f "STRINGREF"))
(targ-op "##string-set!"         (targ-APPLY-simp #t "STRINGSET"))
(targ-op "##string-shrink!"      (targ-APPLY-simp #t "STRINGSHRINK"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##vector8"             (targ-APPLY-vector 'VECTOR8))
(targ-op "##vector8-length"      (targ-APPLY-simp #f "VECTOR8LENGTH"))
(targ-op "##vector8-ref"         (targ-APPLY-simp #f "VECTOR8REF"))
(targ-op "##vector8-set!"        (targ-APPLY-simp #t "VECTOR8SET"))
(targ-op "##vector8-shrink!"     (targ-APPLY-simp #t "VECTOR8SHRINK"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##vector16"            (targ-APPLY-vector 'VECTOR16))
(targ-op "##vector16-length"     (targ-APPLY-simp #f "VECTOR16LENGTH"))
(targ-op "##vector16-ref"        (targ-APPLY-simp #f "VECTOR16REF"))
(targ-op "##vector16-set!"       (targ-APPLY-simp #t "VECTOR16SET"))
(targ-op "##vector16-shrink!"    (targ-APPLY-simp #t "VECTOR16SHRINK"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##struct-ref"          (targ-IFJUMP-APPLY "STRUCTREF"))
(targ-op "##struct-set!"         (targ-APPLY-simp #t "STRUCTSET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##symbol->string"      (targ-APPLY-simp #f "SYMBOLTOSTR"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##keyword->string"     (targ-APPLY-simp #f "KEYWORDTOSTR"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##closure-code"        (targ-APPLY-simp #f "CLOSURECODE"))
(targ-op "##closure-ref"         (targ-APPLY-simp #f "CLOSUREREF"))
(targ-op "##closure-set!"        (targ-APPLY-simp #t "CLOSURESET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##global-var-ref"      (targ-APPLY-simp #f "GLOBALVARREF"))
(targ-op "##global-var-set!"     (targ-APPLY-simp #t "GLOBALVARSET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##make-placeholder"    (targ-APPLY-make-placeholder))
(targ-op "##force"               (targ-APPLY-force))
(targ-op "##void"                (targ-APPLY-simp #f "VOID"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##identity"            (targ-APPLY-identity))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op
  "##c-code"
  (lambda (prim)
    (proc-obj-inlinable-set!
      prim
      (lambda (opnds loc sn)
        (targ-use-all-res)
        (let ((n (length opnds)))
          (if (= n 0)
            (compiler-error
              "\"##c-code\" needs at least one argument")
            (let ((code (car opnds))
                  (args (map (lambda (opnd) (targ-opnd opnd))
                             (cdr opnds))))
              (if (and (obj? code)
                       (string? (obj-val code)))
                (let loop ((i (- n 1))
                           (rev-args (reverse args))
                           (lst1 (list "#define " c-id-prefix "RESULT "
                                       c-id-prefix "temp"
                                       #\newline
                                       (obj-val code)
                                       #\newline))
                           (lst2 (list #\newline
                                       "RESULT" c-id-prefix "#undef ")))
                  (if (null? rev-args)
                    (begin
                      (targ-emit (cons 'APPEND lst1))
                      (if loc
                        (targ-emit (targ-loc loc '("RESULT"))))
                      (targ-emit (cons 'APPEND (reverse lst2))))
                    (let ((arg (car rev-args)))
                      (loop (- i 1)
                            (cdr rev-args)
                            (append
                              (list "#define " c-id-prefix "ARG" i " " arg
                                    #\newline)
                              lst1)
                            (append
                              (list #\newline
                                    i "ARG" c-id-prefix "#undef ")
                              lst2)))))
                (compiler-error
                  "Argument 1 of \"##c-code\" must be a string constant")))))))))

)

(targ-setup-operations)

;------------------------------------------------------------------------------

; Table of procedure specializations

(define (targ-setup-specializations)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "not"         (targ-s "##not"))
(targ-spec "null?"       (targ-s "##null?"))
(targ-spec "eq?"         (targ-s "##eq?"))
(targ-spec "eof-object?" (targ-s "##eof-object?"))

(targ-spec "pair?"       (targ-s "##pair?"))
(targ-spec "procedure?"  (targ-s "##procedure?"))
(targ-spec "vector?"     (targ-s "##vector?"))
(targ-spec "symbol?"     (targ-s "##symbol?"))
(targ-spec "keyword?"    (targ-s "##keyword?"))
(targ-spec "string?"     (targ-s "##string?"))
(targ-spec "char?"       (targ-s "##char?"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "zero?"       (targ-s-a "##fixnum.zero?"     "##flonum.zero?"))
(targ-spec "positive?"   (targ-s-a "##fixnum.positive?" "##flonum.positive?"))
(targ-spec "negative?"   (targ-s-a "##fixnum.negative?" "##flonum.negative?"))
(targ-spec "odd?"        (targ-s-a "##fixnum.odd?"      #f))
(targ-spec "even?"       (targ-s-a "##fixnum.even?"     #f))

(targ-spec "+"           (targ-u-a "##fixnum.+"         "##flonum.+"))
(targ-spec "*"           (targ-u-a "##fixnum.*"         "##flonum.*"))
(targ-spec "-"           (targ-u-a "##fixnum.-"         "##flonum.-"))
(targ-spec "/"           (targ-u-a #f                   "##flonum./"))
(targ-spec "quotient"    (targ-u-a "##fixnum.quotient"  #f))
(targ-spec "remainder"   (targ-u-a "##fixnum.remainder" #f))
(targ-spec "modulo"      (targ-u-a "##fixnum.modulo"    #f))

(targ-spec "="           (targ-s-a "##fixnum.="         "##flonum.="))
(targ-spec "<"           (targ-s-a "##fixnum.<"         "##flonum.<"))
(targ-spec ">"           (targ-s-a "##fixnum.>"         "##flonum.>"))
(targ-spec "<="          (targ-s-a "##fixnum.<="        "##flonum.<="))
(targ-spec ">="          (targ-s-a "##fixnum.>="        "##flonum.>="))

(targ-spec "abs"         (targ-u-a #f                   "##flonum.abs"))
(targ-spec "floor"       (targ-u-a #f                   "##flonum.floor"))
(targ-spec "ceiling"     (targ-u-a #f                   "##flonum.ceiling"))
(targ-spec "truncate"    (targ-u-a #f                   "##flonum.truncate"))
(targ-spec "round"       (targ-u-a #f                   "##flonum.round"))
(targ-spec "exp"         (targ-u-a #f                   "##flonum.exp"))
(targ-spec "log"         (targ-u-a #f                   "##flonum.log"))
(targ-spec "sin"         (targ-u-a #f                   "##flonum.sin"))
(targ-spec "cos"         (targ-u-a #f                   "##flonum.cos"))
(targ-spec "tan"         (targ-u-a #f                   "##flonum.tan"))
(targ-spec "asin"        (targ-u-a #f                   "##flonum.asin"))
(targ-spec "acos"        (targ-u-a #f                   "##flonum.acos"))
(targ-spec "atan"        (targ-u-a #f                   "##flonum.atan"))
(targ-spec "sqrt"        (targ-u-a #f                   "##flonum.sqrt"))

(targ-spec "integer->char" (targ-s "##fixnum.->char"))
(targ-spec "char->integer" (targ-s "##fixnum.<-char"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "char=?"     (targ-s "##char=?"))
(targ-spec "char<?"     (targ-s "##char<?"))
(targ-spec "char>?"     (targ-s "##char>?"))
(targ-spec "char<=?"    (targ-s "##char<=?"))
(targ-spec "char>=?"    (targ-s "##char>=?"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "cons"       (targ-s "##cons"))
(targ-spec "set-car!"   (targ-u "##set-car!"))
(targ-spec "set-cdr!"   (targ-u "##set-cdr!"))
(targ-spec "car"        (targ-u "##car"))
(targ-spec "cdr"        (targ-u "##cdr"))
(targ-spec "caar"       (targ-u "##caar"))
(targ-spec "cadr"       (targ-u "##cadr"))
(targ-spec "cdar"       (targ-u "##cdar"))
(targ-spec "cddr"       (targ-u "##cddr"))
(targ-spec "caaar"      (targ-u "##caaar"))
(targ-spec "caadr"      (targ-u "##caadr"))
(targ-spec "cadar"      (targ-u "##cadar"))
(targ-spec "caddr"      (targ-u "##caddr"))
(targ-spec "cdaar"      (targ-u "##cdaar"))
(targ-spec "cdadr"      (targ-u "##cdadr"))
(targ-spec "cddar"      (targ-u "##cddar"))
(targ-spec "cdddr"      (targ-u "##cdddr"))
(targ-spec "caaaar"     (targ-u "##caaaar"))
(targ-spec "caaadr"     (targ-u "##caaadr"))
(targ-spec "caadar"     (targ-u "##caadar"))
(targ-spec "caaddr"     (targ-u "##caaddr"))
(targ-spec "cadaar"     (targ-u "##cadaar"))
(targ-spec "cadadr"     (targ-u "##cadadr"))
(targ-spec "caddar"     (targ-u "##caddar"))
(targ-spec "cadddr"     (targ-u "##cadddr"))
(targ-spec "cdaaar"     (targ-u "##cdaaar"))
(targ-spec "cdaadr"     (targ-u "##cdaadr"))
(targ-spec "cdadar"     (targ-u "##cdadar"))
(targ-spec "cdaddr"     (targ-u "##cdaddr"))
(targ-spec "cddaar"     (targ-u "##cddaar"))
(targ-spec "cddadr"     (targ-u "##cddadr"))
(targ-spec "cdddar"     (targ-u "##cdddar"))
(targ-spec "cddddr"     (targ-u "##cddddr"))

(targ-spec "list"       (targ-s "##list"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "vector"         (targ-s "##vector"))
(targ-spec "vector-length"  (targ-u "##vector-length"))
(targ-spec "vector-ref"     (targ-u "##vector-ref"))
(targ-spec "vector-set!"    (targ-u "##vector-set!"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "string"         (targ-s "##string"))
(targ-spec "string-length"  (targ-u "##string-length"))
(targ-spec "string-ref"     (targ-u "##string-ref"))
(targ-spec "string-set!"    (targ-u "##string-set!"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "##structure-ref"  (targ-u "##struct-ref"))
(targ-spec "##structure-set!" (targ-u "##struct-set!"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "symbol->string"  (targ-s "##symbol->string"))
(targ-spec "keyword->string" (targ-s "##keyword->string"))

(targ-spec "touch"          (targ-s "##force"))
(targ-spec "force"          (targ-s "##force"))
(targ-spec "void"           (targ-s "##void"))

(targ-spec "##eqv?"         (targ-s-**eqv?))

(targ-spec "call-with-current-continuation"
                            (targ-s "##call-with-current-continuation"))

)

(targ-setup-specializations)

;------------------------------------------------------------------------------
